#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue May  9 11:11:15 2017

@author: msuzen

%load_ext autoreload
%autoreload 2

"""

import unittest
from bristol.ensembles import Circular
import numpy as np
    
class test_eigen_circular_ensemble(unittest.TestCase):

      epsilon = 1e-9

      def test_eigen_circular_ensemble_01(self):
          mseeds  = [2963416, 235124, 786134]
          ce      = Circular()
          e_cue   = ce.eigen_circular_ensemble(
                                               seeds=mseeds,
                                               N=5,
                                               nchunks=3,
                                               ensemble='CUE', 
                                               parallel=False
                                              )  
          e_coe   = ce.eigen_circular_ensemble(
                                               seeds=mseeds, 
                                               N=5,
                                               nchunks=3,
                                               ensemble='COE', 
                                               parallel=False
                                              )    
          e_cse   = ce.eigen_circular_ensemble(
                                               seeds=mseeds, 
                                               N=5,
                                               nchunks=3,
                                               ensemble='CSE', 
                                               parallel=False
                                              )
          e_cse2  = ce.eigen_circular_ensemble(
                                               seeds=mseeds, 
                                               N=5,
                                               nchunks=3,
                                               ensemble='CSE', 
                                               parallel=False,
                                               adir="upper"
                                              )
          n_cue   = np.imag(e_cue['c_eigen']).sum()
          n_coe   = np.imag(e_coe['c_eigen']).sum()
          n_cse   = np.imag(e_cse['c_eigen']).sum()
          n_cse2  = np.imag(e_cse2['c_eigen']).sum()
          self.assertTrue(n_cue-62.284496117511054 < self.epsilon)
          self.assertTrue(n_coe+21.810798433338746 < self.epsilon)
          self.assertTrue(np.abs(n_cse) < self.epsilon)
          self.assertTrue(np.abs(n_cse2) < self.epsilon)
          
          
      def test_eigen_circular_ensemble_02(self):
          mseeds  = [2963416, 235124, 786134]
          ce      = Circular()
          e_cue   = ce.eigen_circular_ensemble(
                                               seeds=mseeds,
                                               N=5,
                                               nchunks=3,
                                               ensemble='CUE', 
                                               parallel=True
                                              )  
          e_coe   = ce.eigen_circular_ensemble(
                                               seeds=mseeds, 
                                               N=5,
                                               nchunks=3,
                                               ensemble='COE', 
                                               parallel=True
                                              )    
          e_cse   = ce.eigen_circular_ensemble(
                                               seeds=mseeds, 
                                               N=5,
                                               nchunks=3,
                                               ensemble='CSE', 
                                               parallel=True
                                              )
          e_cse2  = ce.eigen_circular_ensemble(
                                               seeds=mseeds, 
                                               N=5,
                                               nchunks=3,
                                               ensemble='CSE', 
                                               parallel=True,
                                               adir="upper"
                                              )
          n_cue   = np.imag(e_cue['c_eigen']).sum()
          n_coe   = np.imag(e_coe['c_eigen']).sum()
          n_cse   = np.imag(e_cse['c_eigen']).sum()
          n_cse2  = np.imag(e_cse2['c_eigen']).sum()
          self.assertTrue(n_cue-62.284496117511054 < self.epsilon)
          self.assertTrue(n_coe+21.810798433338746 < self.epsilon)
          self.assertTrue(np.abs(n_cse)  < self.epsilon) 
          self.assertTrue(np.abs(n_cse2) < self.epsilon)



"""

     cPSE related functionality
     


"""


import numpy as np
import sys
import bristol
from bristol.spectral import Ergodicity
import json
from itertools import cycle

ergo = Ergodicity()

import torchvision
import torchvision.models as models
import torch

def get_layer_matrix_set(pmodel):
    """
    
    Return layer matrix set of a given pre-trained model 
    
    Input
    
    pmodel : pytorch torchvision pre-trained model
    
    Returns:
    
    A tuple (A_set, A_set_N, A_set_types)
        A_set       : A list of 2D np-array, weight matrices
        A_set_N     : Shape of NxN matrices. 
        A_set_types : Layer type, pytorch object type that is
                      extracted as 2D weight matrix.
    
    """
    A_set = []
    A_set_N = []
    A_set_types = []
    for x in pmodel.modules():
        type_mod = str(type(x))  # module/method name
        if "torch.nn.modules" in type_mod:
            try:
                layer_weights = torch.Tensor(x.weight)
                shape_layer = list(layer_weights.shape)
                len_shape = len(shape_layer)
                if len_shape >= 2:
                    N = shape_layer[0]
                    M = np.prod(shape_layer[1:])
                    if N > 1 and M > 1:
                        X = layer_weights
                        Ap = np.array(X.reshape(N, M).detach().numpy())
                        A = np.matmul(Ap, np.transpose(Ap))
                        A_set.append(A)
                        A_set_N.append(A.shape)
                        A_set_types.append(type_mod)
            except:
                pass
    return (A_set, A_set_N, A_set_types)


def get_eigenvals_layer_matrix_set(A_set):
    """
    
    Compute eigenvalues of given set of matrices
    
    Input: 
    
    A_set : list of 2D ndarrays, square real 
    
    Output
    eigenvals_set : List of list of eigenvalues
    
    
    """
    eigenvals_set = []
    for A in A_set:
        eigen_values = np.linalg.eigvals(np.matmul(A, np.transpose(A)))
        eigenvals_set.append(eigen_values)
    return eigenvals_set

def list2plist(lst, upper_bound):
    """
    
    Given list lst ans upper_bound.
    Return period_lst, cycle. 
    
    """
    pool = cycle(lst)
    c = 1
    lst_period = []
    for item in pool:
        c = c + 1
        if isinstance(item, np.complex64):
            item = item.real  # catch for numerical small-unstable numbers
        lst_period.append(abs(item))
        if c > upper_bound:
            break
    return lst_period


def eigenvals_set_to_periodic(layer_eigens):
    """
    
    Layer matrix set eigenvalues to periodic set
    
    """
    upper_bound = np.max([len(e) for e in layer_eigens])
    eset_period = [list2plist(e, upper_bound) for e in layer_eigens]
    return eset_period

def d_layers_pse(eset_per):
    """
    
    Progression of D_layers given periodic set

    Ex:
    from bristol import cPSE
    netname = 'vgg11'
    pmodel = getattr(models, netname)(pretrained=True)
    print(type(pmodel))
    A_t = cPSE.get_layer_matrix_set(pmodel)
    eset = cPSE.get_eigenvals_layer_matrix_set(A_t[0])
    eset_per = cPSE.eigenvals_set_to_periodic(eset)
    d_layers = cPSE.d_layers_pse(eset_per)
    
    """
    nlayers = len(eset_per) - 1  # minus 1 for the last layer
    print(nlayers)
    N = len(eset_per[0])
    D_layer = []
    for l in np.arange(1, nlayers):
        eigen_l = np.ravel(np.array(eset_per[0:l]))
        l1 = l + 1
        eigen_l1 = np.ravel(np.array(eset_per[0:l1]))
        omega_l = ergo.thirumalai_mountain(eigen_l, l, N)
        omega_l1 = ergo.thirumalai_mountain(eigen_l1, l1, N)
        dl = ergo.kl_distance_symmetric(omega_l, omega_l1)
        D_layer.append(dl)
    return D_layer


def cpse_measure(pmodel):
    """
    Given torch model object pmodel return 
    pse on layers and mean log pse Cascading PSE
    (d_layers, cpse) : d_layers vector and real number cpse
     
    netname = 'vgg11'
    pmodel = getattr(models, netname)(pretrained=True)
    (d_layers, cpse) = cPSE.cpse(pmodel)
     
    """
    A_t = get_layer_matrix_set(pmodel)
    eset = get_eigenvals_layer_matrix_set(A_t[0])
    eset_per = eigenvals_set_to_periodic(eset)
    d_layers = d_layers_pse(eset_per)
    return d_layers, np.mean(np.log10(d_layers))

def cpse_measure_vanilla(matrices):
    """

    Given list of weight matrices.
    pse on layers and mean log pse Cascading PSE
    (d_layers, cpse) : d_layers vector and real number cpse

    np.random.seed(42)
    matrices = [np.random.normal(size=(64,64)) for _ in range(10)]
    (d_layers, cpse) = cPSE.cpse_measure_vanilla(matrices)

    """
    eset = get_eigenvals_layer_matrix_set(matrices)
    eset_per = eigenvals_set_to_periodic(eset)
    d_layers = d_layers_pse(eset_per)
    return d_layers, np.mean(np.log10(d_layers))



import unittest
import numpy as np
from bristol import cPSE
import torchvision.models as models

class test_cPSE(unittest.TestCase):

      epsilon = 1e-9

      def test_dlayer_vgg11_01(self):
          netname = 'vgg11'
          pmodel = getattr(models, netname)(pretrained=True)
          d_layers, cpse = cPSE.cpse_measure(pmodel)
          delta0 = d_layers[0] - 620.8614145333568
          delta5 = d_layers[5] - 0.4511635231637898
          delta_c = cpse - 0.03596423728640884
          self.assertTrue(delta0 < self.epsilon)
          self.assertTrue(delta5 < self.epsilon)
          self.assertTrue(delta_c< self.epsilon)



"""
   
     Ergodicity Module:
      

"""

import numpy as np


class Ergodicity:
    def __init__(self):
        pass

    def spectral_density(self, c_eigen, ensemble_size, N, delta_rad=0.2):
        """
             
           Compute spectral density
    
    
           Given set of eigenvalues e_i, compute histogram of 
           Angle(e_i) from -Pi to Pi at a given spacing.
    
           Input
           c_eigen        set of eigenvalues as an np array.
           ensemble_size  number of ensembles used, this is used to 
                          scale the resulting spectrum.
           delta_rad      spacing to use in getting the density, 
                          defaults to 0.2 radians. Used only for complex
    
           Output
           A density in two dimensional numpy array, with bin centres in the 
           first column and the density in the second column.
    
          Example:
              
            import numpy as np
            from bristol.ensembles import Circular
            ce            = Circular()
            mseed         = 2963416
            # 5, 10x10 matrices from CUE
            N             = 10
            ensemble_size = 5
            e_cue         = ce.eigen_circular(
                                              N=N,
                                              ensemble='CUE', 
                                              set_seed=True,
                                              seed=mseed
                                             )    
            from bristol.spectral import Ergodicity
            ergo     = Ergodicity()
            sdensity = ergo.spectral_density(e_cue, ensemble_size, N)
    
        """
        is_C = True
        try:
            sum_c = sum([np.abs(c.imag) for c in c_eigen])
            if (sum_c < 1e-9):
                is_C = False
        except:
            pass
        if (is_C):
            b_ks = np.arange(-np.pi, np.pi, delta_rad)  # bin edges
            b_ks_centres = b_ks[1:] - delta_rad / 2.0  # bin centres
            rho_ensemble = np.histogram(np.angle(c_eigen), bins=b_ks)
            den = np.column_stack(
                (b_ks_centres, rho_ensemble[0] / float(ensemble_size)))
        if (not is_C):
            hist_v, bin_edge = np.histogram(c_eigen)
            bin_centre = bin_edge[:-1] + (bin_edge[1] - bin_edge[0]) / 2.0
            den = np.column_stack((bin_centre, hist_v / float(ensemble_size)))

        return den

    def thirumalai_mountain(self,
                            c_eigen_ensemble,
                            ensemble_size,
                            N,
                            delta_rad=0.2):
        """
         
         Compute TM metric for given set of eigenvalues e_i.
    
    
         Input
          c_eigen_ensemble : set of eigenvalues as a 1D np array
          ensemble_size    : number of ensembles used, this is used to scale the resulting spectrum.
          N                : matrix size used to generate eigenvalues.
          delta_rad        : spacing to use in getting the density, defaults to 0.2 radians.
    
         Output
          Omega, TM metric 1d numpy array.
          
          References:
              Spectral Ergodicity in Deep Learning Architectures 
              via Surrogate Random Matrices, 
              Mehmet Suezen, Cornelius Weber, Joan J. Cerd{`a}, 
              arXiv:1704.08693
          
          Example: 
            # On Ipython
            %load_ext autoreload
            %autoreload 2
         
            import numpy as np
            from bristol.ensembles import Circular
            ce            = Circular()
            mseed         = 2963416
            # 5, 10x10 matrices from CUE
            N             = 10
            cSize         = 2
            nchunks       = 5
            ensemble_size = cSize*nchunks
            eseeds        = [978712, 34687, 43124, 67831, 1234]
            e_cue         = ce.eigen_circular_ensemble(
                                                       N=N,
                                                       ensemble='CUE', 
                                                       seeds=eseeds,
                                                       cSize=cSize,
                                                       nchunks=nchunks
                                                      )    
            c_eigen_ensemble = e_cue['c_eigen']
            from bristol.spectral import Ergodicity
            ergo     = Ergodicity()
            tm       = ergo.thirumalai_mountain(
                                                c_eigen_ensemble, 
                                                ensemble_size, 
                                                N
                                               )
    
        """
        sden_ensemble = self.spectral_density(c_eigen_ensemble, ensemble_size,
                                              N, delta_rad)
        omega = np.zeros(sden_ensemble.shape[0])
        for i in range(ensemble_size):
            ix = np.arange(0, N) + N * i
            sden_spec = self.spectral_density(c_eigen_ensemble[ix], 1, N,
                                              delta_rad)
            omega = np.power(
                (sden_spec[:, 1] - sden_ensemble[:, 1]), 2) + omega
        return omega / ensemble_size / N

    def kl_distance_symmetric(self, Nk, Nk_minus, shift=1e-9):
        """
    
        Compute Kullback-Leibler Divergence in two directions. 
    
    
        Input:
         Nk, Nk_minus     Two 1D numpy arrays 
         shift            Epsilon shift distribution upwards, to avoid for zeros, defaults to 1e-9. 
    
        Output:
          Distance from KL divergence, floating number.
          
        Example:
            # On Ipython
            %load_ext autoreload
            %autoreload 2
            import numpy as np
            from bristol.spectral import Ergodicity
            ergo     = Ergodicity()
            np.random.seed(1235)
            Nk       = np.random.random(100)
            Nk_minus = np.random.random(100)
            ergo.kl_distance_symmetric(Nk,Nk_minus)
    
        """
        KL_k = np.sum(Nk * np.log2((Nk + shift) / (Nk_minus + shift)))
        KL_k_minus = np.sum(Nk_minus * np.log2(
            (Nk_minus + shift) / (Nk + shift)))
        return ((KL_k + KL_k_minus))

    def approach_se(self, Ns, ensemble_size, eigen_data, delta_rad=0.2):
        """
     
         Approach to spectral ergodicity Consecutive 
         
         Parameters
         ----------
         Ns            : Matrix sizes used in a list
         ensemble_size : Number of matrices used
         eigen_data    : Dictionary with key entries from Ns 
                         and values are outputs from 
                         'eigen_circular_ensemble', see example
                         
         References:
         
         Spectral Ergodicity in Deep Learning Architectures 
         via Surrogate Random Matrices, 
         Mehmet Suezen, Cornelius Weber, Joan J. Cerd{`a}, 
         arXiv:1704.08693 
         
         Example:
         
          import numpy as np     
          from bristol.ensembles import Circular
          ce         = Circular()
          mseed      = [123,125,124]       
          Ns         = [5, 10, 15]
          cSize      = 2
          nChunks    = 3
          eigen_data5 = ce.eigen_circular_ensemble(
                                                   Ns[0], 
                                                   cSize=cSize, 
                                                   nchunks=nChunks, 
                                                   seeds=mseed, 
                                                   parallel=False
                                                  )
         eigen_data10 = ce.eigen_circular_ensemble(
                                                   Ns[1], 
                                                   cSize=cSize, 
                                                   nchunks=nChunks, 
                                                   seeds=mseed, 
                                                   parallel=False
                                                  )
         eigen_data15 = ce.eigen_circular_ensemble(
                                                   Ns[2], 
                                                   cSize=cSize, 
                                                   nchunks=nChunks, 
                                                   seeds=mseed, 
                                                   parallel=False
                                                  )
         ensemble_size = cSize*nChunks 
         eigen_data = {
                       "N5":eigen_data5, 
                       "N10":eigen_data10,
                       "N15":eigen_data15
                      }
         from bristol.spectral import Ergodicity
         ergo = Ergodicity()
         Dse  = ergo.approach_se(
                                 Ns=Ns,
                                 ensemble_size=ensemble_size, 
                                 eigen_data=eigen_data 
                                )
        
        """
        Dse = []
        for i in range(1, len(Ns)):
            c_eigen_ensemble = eigen_data["N" + str(Ns[i])]['c_eigen']
            Nk = self.thirumalai_mountain(c_eigen_ensemble, ensemble_size,
                                          Ns[i], delta_rad)
            c_eigen = eigen_data['N' +
                                 str(Ns[i - 1])]['c_eigen']  # All eigenvalues
            Nk_minus = self.thirumalai_mountain(c_eigen, ensemble_size,
                                                Ns[i - 1], delta_rad)
            Dse.append(self.kl_distance_symmetric(Nk, Nk_minus))
        return (Dse)



import unittest
from bristol.ensembles import Circular
from bristol.spectral import Ergodicity
import numpy as np

class test_thirumalai_mountain(unittest.TestCase):

      epsilon = 1e-9

      def test_test_thirumalai_mountain_01(self):
          ce            = Circular()
          mseed         = 2963416
          # 5, 10x10 matrices from CUE
          N             = 10
          cSize         = 2
          nchunks       = 5
          ensemble_size = cSize*nchunks
          eseeds        = [978712, 34687, 43124, 67831, 1234]
          e_cue         = ce.eigen_circular_ensemble(
                                                     N=N,
                                                     ensemble='CUE', 
                                                     seeds=eseeds,
                                                     cSize=cSize,
                                                     nchunks=nchunks
                                                    )    
          c_eigen_ensemble = e_cue['c_eigen']
          ergo     = Ergodicity()
          tm       = ergo.thirumalai_mountain(
                                              c_eigen_ensemble, 
                                              ensemble_size, 
                                              N
                                             )
          self.assertTrue(tm.shape == (31,))
          self.assertTrue(tm[6:11].sum()-0.104 < self.epsilon)



import unittest
from bristol.ensembles import Circular
import numpy as np

class test_unit_symplectic(unittest.TestCase):

      epsilon = 1e-9

      def test_unit_symplectic1(self):
           N   = 4
           ce  = Circular()
           u_l = ce.unit_symplectic(N) 
           u_u = ce.unit_symplectic(N, adir='upper')
           r_l = u_l[1,0] + u_l[3,2] + u_l[5,4] + u_l[7,6] + 4.0
           r_u = u_l[1,0] + u_l[3,2] + u_l[5,4] + u_l[7,6] - 4.0
           self.assertTrue(r_l < self.epsilon)
           self.assertTrue(r_u < self.epsilon)

      def test_unit_symplectic2(self):
           N   = 3 
           ce  = Circular()
           u_l = ce.unit_symplectic(N) 
           u_u = ce.unit_symplectic(N, adir='upper')
           self.assertTrue(np.sum(u_l+u_u) < self.epsilon)



"""
   
     Generating Circular Ensembles


"""

import numpy as np
import multiprocessing as mp
from builtins import map
from functools import partial

def _n_eigen_circular2(seed, N, size, ensemble='CUE',
                       adir='lower', set_seed=False):
        """
        This is a wrapper for _n_eigen_circular, so `seed` comes
        as first argument.
        

        """
        ce = Circular()
        return(ce._n_eigen_circular(N, size, ensemble=ensemble,
                     adir='lower', set_seed=set_seed, seed=seed))


class Circular:

    def __init__(self):
       pass

    def unit_anti(self, N=2, adir='lower'):
        """
        
        Generate antisymmetric unit matrix of size NxN.
        
        
        params:
        N      Size of rectangular NxN ndarray, defaults to 2.
        adir  Direction of Antisymmetry, 'upper' or 'lower' 
          triangular,  defaults to 'lower'.
        
        output:
        z      Antisymmetric unit matrix, NxN ndarray
        
        Example:
        from bristol.ensembles import Circular
        ce = Circular()
        z = ce.unit_anti()         # Generates 2x2, lower anti-symmetric.
        z = ce.unit_anti(adir='upper')  # upper anti-symmetric
        
        """
        z = np.ones([N, N])
        if adir == 'upper':
            upper_ix = np.triu_indices(N)
            z[upper_ix] = -1*z[upper_ix]
            np.fill_diagonal(z, 0)
        if adir == 'lower':
            lower_ix = np.tril_indices(N)
            z[lower_ix]= -1*z[lower_ix]
            np.fill_diagonal(z, 0)
        return z

    def unit_symplectic(self, N, adir='lower'):
        """
        
        Generate unit symplectic matrix 2Nx2N.
        
        
        params:
        N      Base size of rectangular array 2Nx2N ndarray.
        adir  Direction of Antisymmetry, 'upper' or 'lower' triangular,  defaults to 'lower'.
        
        output:
        2Nx2N unit symplectic matrix
        
        Example:
        from bristol.ensembles import Circular
        ce = Circular()
        z  = ce.unit_symplectic(4)         # Generates 8x8 unit symplectic, upper anti-symmetric.
        z  = ce.unit_symplectic(4,adir='upper')  # upper anti-symmetric
        
        """
        z = self.unit_anti(2, adir=adir)
        return np.kron(np.eye(N), z).astype(int)

    def gen_cue(self, N, set_seed=False, seed=42391):
        """
        
        Generate random matrix Circular Unitary Ensemble (CUE)

        Review/reporter: C. Garbers
        
        params:
        N          Size of rectangular array NxN.
        set_seed  Option to pass seed, defaults to False, no seed set.
        seed      If set_seed is set, seed value will be used, defaults to 42391.

        output:
        NxN matrix in  Circular Unitary Ensemble (CUE)
       
        References:
        * M V Berry and Pragya Shukla, New Journal of Physics 15 (2013) 013026 

        Example:
         
        from bristol.ensembles import Circular
        ce     = Circular()
        mseed  = 2963416
        Hcue0  = ce.gen_cue(8, seed=mseed, set_seed=True)
        Hcue1  = ce.gen_cue(8, seed=mseed, set_seed=True)    
        n0     = float(np.imag(Hcue0[0,:].sum()))
        n1     = float(np.real(Hcue1[6,:].sum()))
        n0, n1

                   
        """
        if set_seed:
            np.random.seed(seed)
        f_uni   = lambda n: list(np.random.uniform(0, np.pi *2, n))
        G       = list(map(lambda theta: np.cos(theta)+np.sin(theta)*1j, f_uni(N)))
        A       = np.random.random((N, N))
        B       = np.random.random((N, N))
        H       = 0.5 * (A+B*1j+np.transpose(A)-np.transpose(B)*1j)
        E, U    = np.linalg.eig(H)
        Hcue    = G * U
        return Hcue

    def gen_coe(self, N, set_seed=False, seed=42391):
       """

        Generate random matrix Circular Orthogonal Ensemble (COE)


        params:
        N          Size of rectangular array NxN.
        set_seed  Option to pass seed, defaults to False, no seed set.
        seed      If set_seed is set, seed value will be used, defaults to 42391.

        output:
        NxN matrix in  Circular Orthogonal Ensemble (COE)

        References:
        * M V Berry and Pragya Shukla, New Journal of Physics 15 (2013) 013026

        Example:


        from bristol.ensembles import Circular
        ce     = Circular()
        mseed  = 2963416
        Hcoe0  = ce.gen_coe(8, seed=mseed, set_seed=True)
        Hcoe1  = ce.gen_coe(8, seed=mseed, set_seed=True)
        n0     = float(np.imag(Hcoe0[0,:].sum()))
        n1     = float(np.real(Hcoe1[6,:].sum()))
        n0, n1

       """
       Hcue = self.gen_cue(N,set_seed,seed)
       return Hcue.transpose()*Hcue

    def gen_cse(self, N, set_seed=False, seed=42391, adir='lower'):
       """

        Generate random matrix Circular Symplectic Ensemble (CSE)


        params:
        N          Size of rectangular array NxN.
        set_seed  Option to pass seed, defaults to False, no seed set.
        seed      If set_seed is set, seed value will be used, defaults to 42391.
        adir      Direction of Antisymmetry, 'upper' or 'lower' triangular,  defaults to 'lower'.

        output:
        NxN matrix in  Circular Symlectic Ensemble (CSE)

        References:
        * M V Berry and Pragya Shukla, New Journal of Physics 15 (2013) 013026

        Example:

        from bristol.ensembles import Circular
        ce     = Circular()
        mseed  = 2963416
        Hcse0  = ce.gen_cse(8, seed=mseed, set_seed=True)
        Hcse1  = ce.gen_cse(8, seed=mseed, set_seed=True)
        n0     = float(np.imag(Hcse0[0,:].sum()))
        n1     = float(np.real(Hcse1[6,:].sum()))
        n0, n1

        mseed  = 2963416
        Hcse0  = ce.gen_cse(8, seed=mseed, set_seed=True, adir="upper")
        Hcse1  = ce.gen_cse(8, seed=mseed, set_seed=True, adir="upper")
        n0     = float(np.imag(Hcse0[0,:].sum()))
        n1     = float(np.real(Hcse1[6,:].sum()))
        n0, n1

       """
       Z    = self.unit_symplectic(N, adir=adir)
       Hcue = self.gen_cue(2*N,set_seed,seed)
       return (Z*Hcue.transpose()*Z)*Hcue

    def eigen_circular(self, N, ensemble='CUE', set_seed=False,
                       seed=42391, adir='lower'):
        """

           Generate eigenvalues of a matrix that is a realization from circular 
           random matrix ensemble.
          

           params:
           N     Size of the square random matrix, NxN.
           ensemble  One of the Circular ensemble 'CUE', 'COE', 'CSE'
           set_seed  Option to pass seed, defaults to False, no seed set.
           seed     If set_seed is set, seed value will be used, defaults to 42391.
           adir     Direction of Antisymmetry, 'upper' or 'lower' triangular,  defaults to 'lower'.
           
           output:
           e  eigenvalues as numpy array
         

           Example:
           from bristol.ensembles import Circular
           ce      = Circular()
           mseed   = 2963416    
           e_cue   = ce.eigen_circular(5,  ensemble='CUE', set_seed=True, seed=mseed)
           e_coe   = ce.eigen_circular(10, ensemble='COE', set_seed=True, seed=mseed)
           e_cse   = ce.eigen_circular(15, ensemble='COE', set_seed=True, seed=mseed)
           n_cue   = np.imag(e_cue).sum()
           n_coe   = np.imag(e_coe).sum()
           n_cse   = np.imag(e_cse).sum() 

        """
        if(not ensemble in ['CUE', 'COE', 'CSE']):
            raise Exception("Circular ensemble of \
                     CUE, COE or CSE must \
                     be selected.")
        if ensemble == 'CUE':
            H     = self.gen_cue(N,seed=seed,set_seed=set_seed)
        elif ensemble == 'COE':
            H     = self.gen_coe(N,seed=seed,set_seed=set_seed)
        elif ensemble == 'CSE':
            H     = self.gen_cse(N,seed=seed,set_seed=set_seed,adir=adir)
        e, u = np.linalg.eig(H)
        return(e)

    def _n_eigen_circular(self, N, size, ensemble='CUE',
                          adir='lower', set_seed=False, seed=9876):
        """

        Compute eigenvalues of a given circular ensemble, 
        by generating `size` number of matrices. A choice
        of seed is generated if set_seed is False.

        
        params:
        N          Size of the rectangular matrix, NxN.
        size       Number of random matrices to generate.
        ensemble   One of the circular ensemble 'CUE', 'COE', 'CSE'
        set_seed   Option to pass seed, defaults to False, no seed set.
        seed       If set_seed is set, seed value will be used, defaults to 9876.
        adir       Direction of Antisymmetry, 'upper' or 'lower' triangular,  defaults to 'lower'.

        output:
        Dictionary with keys `local_seed` an integer, and numpy array of eigenvalues
        for each matrix in key `c_eigen`.
       
        Example:
        res = _n_eigen_circular(10, 10) 

        """
        local_seed = np.random.choice(1000000, 1)
        np.random.seed(local_seed[0])
        if set_seed:
           np.random.seed(seed)
           local_seed = seed
        c_eigen = np.array([self.eigen_circular(N, ensemble=ensemble, set_seed=set_seed,
                                   seed=local_seed, adir=adir) for i in range(size)])
        return {'local_seed':local_seed, 'c_eigen':c_eigen}


    def eigen_circular_ensemble(self, N, cSize=100, nchunks=4,
                                ensemble='CUE', adir='lower',
                                seeds=list(), parallel=True):
        """
         
        Compute eigenvalues of given circular ensemble, in parallel or serial.
        

        params:
        N           Size of the rectangular matrix, NxN.
        cSize      Number of random matrices to generate in a chunk.
        nchunks    number of cSize chunks.
        ensemble   One of the circular ensemble 'CUE', 'COE', 'CSE', defaults to 'CUE'
        adir       Direction of Antisymmetry, 'upper' or 'lower' triangular,  defaults to 'lower'.
        seeds      List of integer to use in random seed in every chunk.
        parallel   Run in multicore, number of cores as nchunks, defaults to True

        output:
        Dictionary with keys `local_seed` an integer, and numpy array of eigenvalues
        for each matrix in key `c_eigen`.

        Example:

        Parallel and serial runs should produce the same result with the
        same random seed:
        
        from bristol.ensembles import Circular
        ce        = Circular()
        mseed     = [123,125,124,122]       
        rr_serial = ce.eigen_circular_ensemble(
                                               10, 
                                               cSize=4, 
                                               nchunks=4, 
                                               seeds=mseed, 
                                               parallel=False
                                              )
        rr_paralel = ce.eigen_circular_ensemble(
                                                10, 
                                                cSize=4, 
                                                nchunks=4, 
                                                seeds=mseed, 
                                                parallel=True
                                               )
        all(rr_serial['c_eigen'] == rr_paralel['c_eigen']) 


        """
        if len(seeds) != nchunks:
          raise Exception("Seeds vector must be provided for each chunk")
        if(not parallel):
         local_seeds = []
         c_eigen     = np.empty(0)
         for i in range(nchunks):
             res = {}
             res = self._n_eigen_circular(N=N, size=cSize, ensemble=ensemble,
                         adir='lower', set_seed=True,
                         seed=seeds[i])
             local_seeds.append(res['local_seed'])
             c_eigen = np.append(c_eigen, res['c_eigen'])
         return({'local_seeds':local_seeds, 'c_eigen':c_eigen})
        if parallel:
          wrap_f      = partial(_n_eigen_circular2, N=N,
                                size=cSize, ensemble=ensemble,
                                adir='lower', set_seed=True)
          pool        = mp.Pool(processes=nchunks)
          rrp         = pool.map(wrap_f, seeds)
          pool.close()
          pool.join()
          local_seeds = []
          c_eigen      = np.empty(0)
          for j in range(nchunks):
             local_seeds.append(rrp[j]['local_seed'])
             c_eigen = np.append(c_eigen, rrp[j]['c_eigen'])
          return {'local_seeds':local_seeds, 'c_eigen':c_eigen,
              'matrix_size':N, 'number_of_matrices':nchunks*cSize}




"""
   
     Data Generation Module
     

"""
import numpy as np
from bristol.ensembles import Circular

class Generate: 

      def __init_(self):
          pass

      def spectra_ce(
                     range_N=[64, 128],
                     cSize=5,
                     nchunks=2,
                     seeds=[997123, 1091645],
                     ensemble='all',
                     parallel=False,
                    ):
          """

          Generate eigenvalues of matrices of different size
          drawn from a circular ensemble.


           params:
           range_N   set of Size of the rectangular matrix, NxN.
           cSize     Number of random matrices to generate in a chunk.
           nchunks   number of cSize chunks.
           ensemble  One of the circular ensemble 
                     'CUE', 'COE', 'CSE', defaults to 'all'
           seeds     List of integer to use in random seed 
                     in every chunk.
           parallel  Run in multicore, number of cores as nchunks, 
                     defaults to False

          output:
          Dictionary of dictionaries. 
          * Highest level keys for ensemble `CUE`, `COE` or `CSE`.
          * Then, dictionary with keys describing size, such as 'N16. 
          * Values will be an another dictionary
            with keys `local_seed` integers for seeds 
            used in the chunk, `c_eigen` 
            numpy array of eigenvalues, `N` matrix size, 
            `M` number of matrices used to generate `c_eigen`. Note
            that `c_eigen` will be length NxM.
 
            Example:

          """
          ce      = Circular()
          data_ce = {
                     'CUE':{},
                     'COE':{},
                     'CSE':{}
                    }
          for N in range_N: 
              if(ensemble in ['CUE','all']):
                e_dict = ce.eigen_circular_ensemble(
                                                    N=N,
                                                    cSize=cSize,
                                                    nchunks=nchunks,
                                                    ensemble='CUE',
                                                    seeds=seeds,
                                                    parallel=parallel
                                                   )
                data_ce['CUE']['N'+str(N)] = e_dict
              if(ensemble in ['COE','all']):
                e_dict = ce.eigen_circular_ensemble(
                                                    N=N,
                                                    cSize=cSize,
                                                    nchunks=nchunks,
                                                    ensemble='COE',
                                                    seeds=seeds,
                                                    parallel=parallel
                                                   )
                data_ce['COE']['N'+str(N)] = e_dict
              if(ensemble in ['CSE','all']):
                e_dict = ce.eigen_circular_ensemble(
                                                    N=N,
                                                    cSize=cSize,
                                                    nchunks=nchunks,
                                                    ensemble='CSE',
                                                    seeds=seeds,
                                                    parallel=parallel
                                                   )
                data_ce['COE']['N'+str(N)] = e_dict
          return data_ce



import unittest
from bristol.ensembles import Circular
import numpy as np

class test_gen_cue(unittest.TestCase):

      epsilon = 1e-9

      def test_gen_cue_01(self):
           ce    = Circular()
           mseed = 2963416
           Hcue0 = ce.gen_cue(8, seed=mseed, set_seed=True)
           Hcue1 = ce.gen_cue(8, seed=mseed, set_seed=True)    
           n0    =  float(np.imag(Hcue0[0,:].sum()))+0.1445812268759293
           n1    =  float(np.real(Hcue1[6,:].sum()))-0.0445623361854815
           self.assertTrue(n0 < self.epsilon)
           self.assertTrue(n1 < self.epsilon)



import unittest
from bristol.ensembles import Circular

class test_unit_anti(unittest.TestCase):

      epsilon = 1e-9

      def test_unit_anti1(self):
           ce  = Circular()
           z_l = ce.unit_anti() 
           z_u = ce.unit_anti(adir='upper')
           self.assertTrue((z_l[1,0] + z_l[0,1]) < self.epsilon)
           self.assertTrue((z_l[1,0] + 1.0) < self.epsilon)
           self.assertTrue((z_u[0,1] + z_u[0,1]) < self.epsilon)
           self.assertTrue((z_u[0,1] + 1.0) < self.epsilon)

      def test_unit_anti2(self):
           N   = 4
           ce  = Circular()
           z_l = ce.unit_anti(N)
           z_u = ce.unit_anti(N, adir='upper')
           self.assertTrue(z_l[3,].sum()+3.0 < self.epsilon)
           self.assertTrue(z_u[3,].sum()-3.0 < self.epsilon)



import unittest
from bristol.ensembles import Circular
from bristol.spectral import Ergodicity

class test_spectral_density(unittest.TestCase):

      epsilon = 1e-9

      def test_spectral_density_01(self):
           N        = 10 
           ce       = Circular()
           mseed    = 2963416
           e_cue    = ce.eigen_circular(
                                        N=N,
                                        ensemble='CUE', 
                                        set_seed=True,
                                        seed=mseed
                                       )
           ergo     = Ergodicity()
           sdensity = ergo.spectral_density(e_cue, ensemble_size=1, N=N)
           self.assertTrue(sdensity.shape == (31,2))
           self.assertTrue(sdensity[5:15,1].sum()-4.0 < self.epsilon)
           self.assertTrue(sdensity[15:27,0].sum()-12.700888156922527 < self.epsilon)



"""
   
     Statistics Helper Methods

"""


class histogram:

    def __init__(self):
       print("histogram")    

    def h_freedman_diaconis(self, x):
        """

          Find optimal bin size for histogram of x vector.
          

          param:
          x    numpy array

          output:
          h    return estimated bin size in freedman_diaconis

          Example:
          x = np.random.random(100)
          h_freedman_diaconis(x)

        """
        q75, q25 = np.percentile(x,[75,25])
        iqr      = q75 - q25
        n3       = np.power(len(x),-1.0/3.0)
        return(2*iqr*n3)



from setuptools import setup

fp = open('./bristol/version.py')
exec(fp.read())

with open('README.md') as f:
    long_description = f.read()

setup(
      name='bristol',
      version=__version__,
      description="Parallel random matrix tools and random matrix theory deep learning applications. Generate matrices from Circular Unitary Ensemble (CUE), Circular Ortogonal Ensemble (COE) and Circular Symplectic Ensemble (CSE). Additional spectral analysis utilities are also implemented, such as computation of spectral density and spectral ergodicity for complexity of deep learning architectures.",
      long_description=long_description,
      long_description_content_type="text/markdown",
      url="https://github.com/msuzen/bristol",
      author='M.Suzen',
      author_email='suzen@acm.org',
      license= 'GPL-3',
      packages=['bristol'],
      install_requires=[
                        'numpy >= 1.11', 
                        'torch >= 1.3.0', 
                        'torchvision >= 0.4.1a0+d94043a'
                       ],
      test_suite="test",
      zip_safe=False
     )
